import dataclasses
import importlib
import inspect
import pkgutil
import re
import warnings
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, List, Optional, cast

import ipywidgets
import reacton
import reacton.core

import solara
from solara.alias import rv
from solara.util import cwd

autoroute_level_context = solara.create_context(0)
DEBUG = False


DefaultLayout = solara.AppLayout


def source_to_module(path: Path) -> ModuleType:
    fullname = path.stem
    mod = ModuleType(fullname)
    mod.__file__ = str(path)
    if path.suffix == ".py":
        with open(path) as f:
            ast = compile(f.read(), path, "exec")
            exec(ast, mod.__dict__)
    elif path.suffix == ".ipynb":
        import nbformat

        nb: nbformat.NotebookNode = nbformat.read(path, 4)
        with cwd(Path(path).parent):
            for cell_index, cell in enumerate(nb.cells):
                cell_index += 1  # used 1 based
                if cell.cell_type == "code":
                    source = cell.source
                    cell_path = f"{path} input cell {cell_index}"
                    ast = compile(source, cell_path, "exec")
                    exec(ast, mod.__dict__)
    else:
        raise TypeError(f"Only .py and .ipynb supported (not {path})")
    return mod


def count_arguments(f: Callable):
    if isinstance(f, reacton.core.ComponentFunction):
        f = f.f
    sig = inspect.signature(f)
    return len([k for name, k in sig.parameters.items() if (k.default == k.empty) and (k.kind != k.VAR_POSITIONAL) and (k.kind != k.VAR_KEYWORD)])


def arg_cast(args: List[str], f: Callable):
    if isinstance(f, reacton.core.ComponentFunction):
        f = f.f
    sig = inspect.signature(f)
    results = []
    for i, (name, param) in enumerate(sig.parameters.items()):
        if i >= len(args):
            break
            # TODO: can we give a better error message?
            # raise TypeError(f"{f} expected positional argument {param}, but not enought argument supplied")
        annotation = param.annotation
        value = args[i]
        check_optional_types = [str, int, float]
        for check_type in check_optional_types:
            if annotation == Optional[check_type]:
                annotation = check_type
        if annotation == param.empty:
            results.append(value)
        else:
            try:
                results.append(annotation(value))
            except Exception as e:
                raise ValueError(f"Cannot cast argument {name} of {f} to type {annotation} with value {value}") from e
    return results


@solara.component
def RoutingProvider(children: List[reacton.core.Element] = [], routes: List[solara.Route] = [], pathname: str = ""):
    """Wraps the app, adds extra context, like navigation/routing."""
    path, set_path = solara.use_state(pathname, key="solara-context-path")
    nav = solara.Navigator(location=path, on_location=set_path)
    solara.routing._location_context.provide(solara.routing._Location(path, set_path))
    solara.routing.router_context.provide(solara.routing.Router(path, routes=routes, set_path=set_path))

    main = solara.VBox(
        children=[
            nav,
            *children,
        ]
    )

    return main


@solara.component
def RenderPage():
    """Renders the page that matches the route."""
    level_start = solara.use_route_level()
    router = solara.use_context(solara.routing.router_context)
    # we use these to cache script runs that use regular ipywidgets
    modules = solara.use_memo(dict, dependencies=[])
    modules_modified_times = solara.use_memo(dict, dependencies=[])

    if len(router.path_routes) <= level_start:
        with solara.VBox() as main:
            solara.Error(f"Page not found: {router.path}, len(router.path_routes)={len(router.path_routes)} <= level_start={level_start}")
            parent = "/" + "/".join(router.parts[:-1])
            with solara.Link(parent):
                solara.Button(f"Go to parent: {parent}", text=True)
            if DEBUG:
                from reacton.core import pp

                from solara.components.captureoutput import CaptureOutput

                with CaptureOutput():
                    pp(router.routes)
        return main

    level_max = level_start
    layouts = []  # nested layouts
    # find the 'RenderPage' sentinel value to find the deepest level we should render
    for level in range(level_start, len(router.path_routes)):
        # we always level_start the 'package'/'root' layout
        roots = [k for k in router.path_routes_siblings[level] if k.path == "/"]
        if len(roots) == 1:
            root = roots[0]
            if root.layout:
                layouts.append(root.layout)
        route = router.path_routes[level]
        # and, if not the root layout, include the layout for the current route
        if route.path != "/" and route.layout:
            layouts.append(route.layout)
        if route.component is RenderPage:
            level_max = level
    route_current = router.path_routes[level_max]
    routes_siblings = router.path_routes_siblings[level_max]
    routes_siblings_index = routes_siblings.index(route_current)
    # if no layouts are found, we use the default layout
    if layouts == []:
        # except when the root routes has no siblings (maybe we should include no children?)
        if len(router.path_routes_siblings[level_start]) == 1:
            layouts = []
        else:
            layouts = [DefaultLayout]
    if route_current.data is None and route_current.module is None:
        return solara.Error(f"Page not found: {router.path}, route does not link to a path or module")

    def wrap_in_layouts(element: reacton.core.Element, layouts):
        for Layout in reversed(layouts):
            element = Layout(children=[element])
        return element

    def get_args(f):
        if len(router.path_routes) < len(router.parts):
            arg_strings = router.parts[len(router.path_routes) :]
            args = arg_cast(arg_strings, f)
            return args
        else:
            return []

    if isinstance(route_current.data, Path):
        path = cast(Path, route_current.data)
        if path.suffix == ".md":
            with solara.HBox() as navigation:
                if routes_siblings_index > 0:
                    prev = routes_siblings[routes_siblings_index - 1]
                    with solara.Link(prev):
                        solara.Button(f"{prev.label}", text=True, icon_name="mdi-arrow-left")
                rv.Spacer()
                if routes_siblings_index < len(routes_siblings) - 1:
                    next = routes_siblings[routes_siblings_index + 1]
                    with solara.Link(next):
                        solara.Button(f"{next.label}", text=True, icon_name="mdi-arrow-right")
            main = solara.Div(
                children=[
                    solara.Title(route_current.label or "No title"),
                    solara.Markdown(path.read_text(), unsafe_solara_execute=True),
                    navigation,
                ]
            )
            main = wrap_in_layouts(main, layouts)
        else:
            main = solara.Error(f"Suffix {path.suffix} not supported")
    else:
        assert route_current.module is not None
        title = route_current.label or "No title"
        title_element = solara.Title(title)
        module = route_current.module
        namespace = module.__dict__
        Page = namespace.get("Page", None)
        # app is for backwards compatibility
        page = namespace.get("page", namespace.get("app"))
        if page is not None:
            if isinstance(page, reacton.core.Element):
                pass  # we are good
            elif isinstance(page, ipywidgets.Widget):
                # If we have a widget, we need to execute this again for each
                # connection, since we cannot share widgets between connections/users.
                # We also cannot tear them down, so we cache the widget based pages.
                # To support hot reload, we manualy need to check the mtimes
                # because the reload support for modules in reloader.py only works
                # for modules.
                if route_current.file is None:
                    page = solara.Error(f"{route_current.path} is not associated with a file")
                else:
                    assert route_current.file is not None
                    if route_current.path not in modules:
                        modules[route_current.path] = source_to_module(route_current.file)
                        modules_modified_times[route_current.path] = route_current.file.stat().st_mtime
                    else:
                        if modules_modified_times[route_current.path] != route_current.file.stat().st_mtime:
                            # out of date, 'reload'
                            modules[route_current.path] = source_to_module(route_current.file)
                            modules_modified_times[route_current.path] = route_current.file.stat().st_mtime
                    page = getattr(modules[route_current.path], "app", None)
                    page = getattr(modules[route_current.path], "page", page)
                    if page is None:
                        page = solara.Error(f"{module} does not have a `Page` component of a `page` element or widget")
            else:
                page = solara.Error(f"{module} page variable not a Solara element or ipywidget")
            main = solara.Div(
                children=[
                    title_element,
                    page,
                ]
            )
            main = wrap_in_layouts(main, layouts)
        elif Page is not None:
            args = get_args(Page)
            main = solara.Div(
                children=[
                    title_element,
                    Page(*args),
                ]
            )
            main = wrap_in_layouts(main, layouts)
        else:
            with DefaultLayout() as main:
                solara.Error(f"{module} does not have a Page component or an app element")
    return main


def get_page(module: ModuleType, required=True):
    page = getattr(module, "Page", None)
    if required and page is None:
        raise NameError(f"No component names 'Page' in module {module}")
    return page


def get_renderable(module: ModuleType, required=False):
    var_names = "app Page page".split()
    for var_name in var_names:
        if not hasattr(module, var_name):
            continue
        entry = getattr(module, var_name)
        return entry
    if required:
        raise NameError(f"No component, element or widget found in module {module} with names: {', '.join(var_names[:-1])} or {var_names[-1]}")


def get_title(module: ModuleType, required=True):
    assert module.__file__ is not None
    name = Path(module.__file__).stem
    if module.__file__.endswith("__init__.py"):
        name = Path(module.__file__).parent.stem
    # if title is a submodule it may shadown the variable in __init__
    # in this case, _title is an escape hatch
    if hasattr(module, "_title") and isinstance(module._title, str):
        title = module._title
    elif hasattr(module, "title") and isinstance(module.title, str):
        title = module.title
    else:
        title_parts = re.split("[\\-_ ]+", name)
        title = " ".join(k.title() for k in title_parts)
    return title


def fix_route(route: solara.Route, new_file: Path) -> solara.Route:
    file = route.file or new_file
    children = fix_routes(route.children, new_file) if route.children else []

    return dataclasses.replace(route, file=file, children=children)


def fix_routes(routes: List[solara.Route], new_file: Path):
    new_routes = []
    for route in routes:
        new_routes.append(fix_route(route, new_file))
    return new_routes


def generate_routes(module: ModuleType) -> List[solara.Route]:
    """Generate routes from a module.

    This is a recursive function that will generate routes for all submodules.

    Note this only support .py files, since Markdown files and Jupyter notebook
    files are not Python modules.


    See [our multipage guide](/docs/howto/multipage#as-a-package) for more details.


    """
    from .server import reload

    assert module.__file__ is not None
    routes = []
    file = Path(module.__file__)

    if module.__file__.endswith("__init__.py"):
        if hasattr(module, "routes"):
            # if routes if provided, use them instead of us generating them
            children = getattr(module, "routes")
            children = fix_routes(children, file)
            return children
        route_order = getattr(module, "route_order", None)
        layout = getattr(module, "Layout", None)
        title = get_title(module)
        children = getattr(module, "routes", [])
        if hasattr(module, "Page"):
            routes.append(solara.Route(path="/", component=RenderPage, data=module, module=module, layout=layout, children=children, label=title, file=file))

        assert module.__file__ is not None
        reload.reloader.watcher.add_file(module.__file__)
        for info in pkgutil.iter_modules([str(Path(module.__file__).parent)]):
            submod = importlib.import_module(module.__name__ + f".{info.name}")
            subfile = Path(submod.__file__) if submod.__file__ is not None else None
            title = get_title(submod)

            if info.ispkg:
                route = solara.Route(info.name, component=RenderPage, children=generate_routes(submod), module=submod, layout=None, label=title)
                # skip empty subpackages
                if len(route.children) == 0:
                    continue
            else:
                # skip empty modules
                if get_renderable(submod) is None:
                    continue
                children = getattr(submod, "routes", [])
                if subfile:
                    children = fix_routes(children, subfile)
                module_layout = getattr(submod, "Layout", None)
                route = solara.Route(info.name, component=RenderPage, module=submod, layout=module_layout, children=children, label=title, file=subfile)
            routes.append(route)
        if route_order:
            lookup = {k.path: k for k in routes}
            for k in route_order:
                if k not in lookup:
                    raise KeyError(f"Route {k!r} listen in route_order not found in {module}")
            routes = [lookup[k] for k in route_order]
            if set(lookup) - set(route_order):
                warnings.warn(f"Some routes are not in route_order: {set(lookup) - set(route_order)}")

    else:
        children = getattr(module, "routes", [])
        children = fix_routes(children, file)
        # children = []
        # single module, single route
        return [solara.Route(path="/", component=RenderPage, data=None, module=module, label=get_title(module), children=children, file=file)]

    return routes


def generate_routes_directory(path: Path) -> List[solara.Route]:
    """Generate routes for a directory.

    This is a recursive function that will generate routes for all
    subdirectories and files in the directory. It will skip any
    files or directories that start with an underscore or a dot.

    Markdown files ending in .md will be rendered as markdown.

    Python files ending in .py, or Jupyter notebooks ending in .ipynb
    will be executed and its `Page` component will be rendered.

    Automatic titles will be [generated as explained in the multipage guide](/docs/howto/multipage).

    """
    from .server import reload

    subpaths = list(sorted(path.iterdir()))
    routes = []
    first = True
    has_index = len([k for k in subpaths if k.name == "__init__"]) > 0
    suffixes = [".py", ".ipynb", ".md"]
    init = path / "__init__.py"
    layout = None
    if init.exists():
        init_module = source_to_module(init)
        layout = getattr(init_module, "Layout", None)

    for subpath in subpaths:
        # only handle directories and recognized file types
        if not (subpath.is_dir() or subpath.suffix in suffixes):
            continue
        if subpath.stem.startswith("_") or subpath.stem.startswith("."):
            continue
        name = subpath.stem
        match = re.match("([0-9\\-_ ]*)(.*)", name)
        if match:
            _prefix, name = match.groups()
        title_parts = re.split("[\\-_ ]+", name)
        title = " ".join(k.title() for k in title_parts)
        if not has_index and first:
            route_path = "/"
        else:
            route_path = "-".join([k.lower() for k in title_parts])
        # used as a 'sentinel' to find the deepest level of the route tree we need to render in 'RenderPage'
        component = RenderPage
        children = []
        module: Optional[ModuleType] = None
        data: Any = None
        module_layout = layout if first else None
        if subpath.suffix == ".md":
            data = subpath
            reload.reloader.watcher.add_file(subpath)
        elif subpath.is_dir():
            children = generate_routes_directory(subpath)
        else:
            reload.reloader.watcher.add_file(subpath)
            module = source_to_module(subpath)
            children = getattr(module, "routes", children)
            children = fix_routes(children, subpath)
            module_layout = getattr(module, "Layout", module_layout)
        first = False
        route = solara.Route(route_path, component=component, module=module, label=title, children=children, data=data, layout=module_layout, file=subpath)
        routes.append(route)
    return routes



import json
import logging
import subprocess
import sys
import warnings
from pathlib import Path
from typing import Optional

import IPython.display
from IPython.display import display

import solara

HERE = Path(__file__).parent
logger = logging.getLogger(__name__)


def getcmdline(pid):
    # for linux
    if sys.platform == "linux":
        with open(f"/proc/{pid}/cmdline", "rb") as f:
            return f.read().split(b"\00")[0].decode("utf-8")
    elif sys.platform == "darwin":
        return subprocess.check_output(["ps", "-o", "command=", "-p", str(pid)]).split(b"\n")[0].split(b" ")[0].decode("utf-8")
    elif sys.platform == "win32":
        return subprocess.check_output(["wmic", "process", "get", "commandline", "/format:list"]).split(b"\n")[0].split(b" ")[0].decode("utf-8")
    else:
        raise ValueError(f"Unsupported platform: {sys.platform}")


def get_server_python_executable(silent: bool = False):
    servers = []
    try:
        from jupyter_server import serverapp

        servers += list(serverapp.list_running_servers())
    except ImportError:
        pass
    try:
        from notebook import notebookapp

        servers += list(notebookapp.list_running_servers())
    except ImportError:
        pass

    pythons = [getcmdline(server["pid"]) for server in servers]
    if len(pythons) == 0:
        python = sys.executable
        if not silent:
            warnings.warn("Could not find servers, we are assuming the server is running under Python executable: %s" % python)
    elif len(pythons) > 1:
        info = "\n\t".join(pythons)
        if sys.executable in pythons:
            python = sys.executable
        else:
            python = pythons[0]
            if not silent:
                warnings.warn("Found multiple find servers:\n%s\nWe are assuming the server is running under Python executable: %s" % (info, python))
    else:
        python = pythons[0]
    return python


libraries_minimal = [
    {"python": "ipyvuetify", "classic": "jupyter-vuetify/extension", "lab": "jupyter-vuetify"},
    {"python": "ipyvue", "classic": "jupyter-vue/extension", "lab": "jupyter-vue"},
]

libraries_extra = [
    {"python": "bqplot", "classic": "bqplot/extension", "lab": "bqplot"},
    {"python": "ipyvolume", "classic": "ipyvolume/extension", "lab": "ipyvolume"},
    {"python": "ipywebrtc", "classic": "jupyter-webrtc", "lab": "jupyter-webrtc"},
    {"python": "ipyleaflet", "classic": "ipyleaflet/extension", "lab": "ipyleaflet"},
]


def check_jupyter(
    server_python: Optional[str] = None,
    silent: bool = False,
    libraries: list = libraries_minimal,
    libraries_extra: list = libraries_extra,
    force: bool = False,
    extra: bool = False,
):
    if solara._using_solara_server():
        return
    try:
        python_executable = server_python or get_server_python_executable(silent)
        if python_executable != sys.executable or force:
            libraries_json = json.dumps(libraries + (libraries_extra if extra else []))
            display(
                IPython.display.Javascript(
                    data="""
                    window.jupyter_python_executable = %r;
                    window.jupyter_widget_checks_silent = %s;
                    window.jupyter_widget_checks_libraries = %s;
                    """
                    % (python_executable, str(silent).lower(), libraries_json)
                )
            )
            display(IPython.display.HTML(filename=str(HERE / "checks.html")))
        else:
            if not silent:
                display(
                    IPython.display.HTML(
                        data="<div>Jupyter server is running under the same Python executable as your kernel, no need to check 👍."
                        "<br> <i>Run solara.check_jupyter(force=True) to force checking.</i></div>"
                    )
                )
    except Exception:
        logger.exception("Could not check jupyter-widgets extensions.")


check_jupyter(silent=True)



import solara

github_url = solara.util.github_url(__file__)


@solara.component
def Page():
    with solara.Sidebar():
        solara.Markdown("This is the sidebar at the home page11!")
    with solara.Card("Home"):
        solara.Markdown("This is the home page")
        solara.Button(label="View source", icon_name="mdi-github-circle", attributes={"href": github_url, "target": "_blank"}, text=True, outlined=True)


@solara.component
def Layout(children):
    route, routes = solara.use_route()
    return solara.AppLayout(children=children)



import logging
import pdb
import sys
import threading
import traceback
import warnings
from typing import MutableMapping
from unittest import mock

import ipykernel.kernelbase
import IPython.display
import ipywidgets

from . import app, reload, settings
from .utils import pdb_guard

logger = logging.getLogger("solara.server.app")


class FakeIPython:
    def __init__(self, context: app.AppContext):
        self.context = context
        self.kernel = context.kernel
        self.display_pub = mock.MagicMock()
        # needed for the pyplot interface of matplotlib
        # (although we don't really support it)
        self.events = mock.MagicMock()

    def enable_gui(self, gui):
        logger.error("ignoring call to enable_gui(%s)", gui)

    def register_post_execute(self, callback):
        # mpl requires this
        pass

    def set_parent(self, *args):
        pass

    def showtraceback(self):
        if settings.main.use_pdb:
            logger.exception("Exception, will be handled by debugger")
            pdb.post_mortem()
        etype, value, tb = sys.exc_info()
        traceback_string = "".join(traceback.format_exception(etype, value, tb))
        logger.error("Uncaught exception: %s", traceback_string)
        msg = {
            "type": "exception",
            "traceback": traceback_string,
        }

        for socket in self.context.control_sockets:
            try:
                socket.send_json(msg)
            except:  # noqa
                # TODO: should we remove it from the list?
                pass


def kernel_instance_dispatch(cls, *args, **kwargs):
    context = app.get_current_context()
    return context.kernel


def kernel_initialized_dispatch(cls):
    try:
        app.get_current_context()
    except RuntimeError:
        return False
    return True


def display_solara(
    *objs,
    include=None,
    exclude=None,
    metadata=None,
    transient=None,
    display_id=None,
    raw=False,
    clear=False,
    **kwargs,
):
    print(*objs)  # noqa


# from IPython.core.interactiveshell import InteractiveShell

# if transient is None:
#     transient = {}
# if metadata is None:
#     metadata = {}
# from IPython.core.display_functions import _new_id

# if display_id:
#     if display_id is True:
#         display_id = _new_id()
#     transient["display_id"] = display_id
# if kwargs.get("update") and "display_id" not in transient:
#     raise TypeError("display_id required for update_display")
# if transient:
#     kwargs["transient"] = transient

# if not objs and display_id:
#     # if given no objects, but still a request for a display_id,
#     # we assume the user wants to insert an empty output that
#     # can be updated later
#     objs = [{}]
#     raw = True

# if not raw:
#     format = InteractiveShell.instance().display_formatter.format

# if clear:
#     clear_output(wait=True)

# for obj in objs:
#     if raw:
#         publish_display_data(data=obj, metadata=metadata, **kwargs)
#     else:
#         format_dict, md_dict = format(obj, include=include, exclude=exclude)
#         if not format_dict:
#             # nothing to display (e.g. _ipython_display_ took over)
#             continue
#         if metadata:
#             # kwarg-specified metadata gets precedence
#             _merge(md_dict, metadata)
#         publish_display_data(data=format_dict, metadata=md_dict, **kwargs)
# if display_id:
#     return DisplayHandle(display_id)


def get_ipython():
    if app.has_current_context():
        context = app.get_current_context()
        our_fake_ipython = FakeIPython(context)
        return our_fake_ipython
    else:
        return None


class context_dict(MutableMapping):
    def _get_context_dict(self) -> dict:
        raise NotImplementedError

    def __delitem__(self, key) -> None:
        self._get_context_dict().__delitem__(key)

    def __getitem__(self, key):
        return self._get_context_dict().__getitem__(key)

    def __iter__(self):
        return self._get_context_dict().__iter__()

    def __len__(self):
        return self._get_context_dict().__len__()

    def __setitem__(self, key, value):
        self._get_context_dict().__setitem__(key, value)


class context_dict_widgets(context_dict):
    def _get_context_dict(self) -> dict:
        context = app.get_current_context()
        return context.widgets


class context_dict_templates(context_dict):
    def _get_context_dict(self) -> dict:
        context = app.get_current_context()
        return context.templates


class context_dict_user(context_dict):
    def __init__(self, name):
        self.name = name

    def _get_context_dict(self) -> dict:
        context = app.get_current_context()
        if self.name not in context.user_dicts:
            context.user_dicts[self.name] = {}
        return context.user_dicts[self.name]


def auto_watch_get_template(get_template):
    """Wraps get_template and adds a file listener for automatic .vue file reloading"""

    def wrapper(abs_path):
        template = get_template(abs_path)
        reload.reloader.watcher.add_file(abs_path)
        return template

    return wrapper


Thread__init__ = threading.Thread.__init__
Thread__run = threading.Thread.run


def WidgetContextAwareThread__init__(self, *args, **kwargs):
    Thread__init__(self, *args, **kwargs)
    self.current_context = None
    try:
        self.current_context = app.get_current_context()
    except RuntimeError:
        logger.debug(f"No context for thread {self}")


def Thread_debug_run(self):
    if self.current_context:
        app.set_context_for_thread(self.current_context, self)
    with pdb_guard():
        Thread__run(self)


_patched = False


def patch():
    global _patched
    if _patched:
        warnings.warn("patch() called twice")
        return
    _patched = True
    __builtins__["display"] = IPython.display.display

    # the ipyvue.Template module cannot be accessed like ipyvue.Template
    # because the import in ipvue overrides it
    template_mod = sys.modules["ipyvue.Template"]
    template_mod.template_registry = context_dict_templates()  # type: ignore
    template_mod.get_template = auto_watch_get_template(template_mod.get_template)  # type: ignore

    # this module also imports get_template
    template_mod_vue = sys.modules["ipyvue.VueTemplateWidget"]
    template_mod_vue.get_template = template_mod.get_template  # type: ignore

    component_mod_vue = sys.modules["ipyvue.VueComponentRegistry"]
    component_mod_vue.vue_component_registry = context_dict_user("vue_component_registry")  # type: ignore
    component_mod_vue.vue_component_files = context_dict_user("vue_component_files")  # type: ignore

    ipywidgets.widget.Widget.widgets = context_dict_widgets()  # type: ignore
    threading.Thread.__init__ = WidgetContextAwareThread__init__  # type: ignore
    threading.Thread.run = Thread_debug_run  # type: ignore
    # on CI we get a mypy error:
    # solara/server/patch.py:210: error: Cannot assign to a method
    #  solara/server/patch.py:210: error: Incompatible types in assignment (expression has type "classmethod[Any]",\
    #                                     variable has type "Callable[[VarArg(Any), KwArg(Any)], Any]")
    # not sure why we cannot reproduce that locally
    ipykernel.kernelbase.Kernel.instance = classmethod(kernel_instance_dispatch)  # type: ignore
    # on CI we get a mypy error:
    # solara/server/patch.py:211: error: Cannot assign to a method
    # solara/server/patch.py:211: error: Incompatible types in assignment (expression has type "classmethod[Any]", variable has type "Callable[[], Any]")
    # not sure why we cannot reproduce that locally
    ipykernel.kernelbase.Kernel.initialized = classmethod(kernel_initialized_dispatch)  # type: ignore
    ipywidgets.widgets.widget.get_ipython = get_ipython
    IPython.get_ipython = get_ipython



import dataclasses
import sys
import threading
from operator import getitem
from typing import Any, Callable, Dict, Generic, Set, Tuple, TypeVar, Union, cast

import react_ipywidgets as react
from reacton.utils import equals

from solara import _using_solara_server

T = TypeVar("T")
TS = TypeVar("TS")
S = TypeVar("S")  # used for state

local = threading.local()
_DEBUG = False


# these hooks should go into react-ipywidgets
def use_sync_external_store(subscribe: Callable[[Callable[[], None]], Callable[[], None]], get_snapshot: Callable[[], Any]):
    _, set_counter = react.use_state(0)

    def force_update():
        set_counter(lambda x: x + 1)

    state = get_snapshot()
    prev_state = react.use_ref(state)

    def update_state():
        prev_state.current = state

    react.use_effect(update_state)

    def on_store_change(_ignore_new_state=None):
        new_state = get_snapshot()
        if not equals(new_state, prev_state.current):
            prev_state.current = new_state
            force_update()

    react.use_effect(lambda: subscribe(on_store_change), [])
    return state


def use_sync_external_store_with_selector(subscribe, get_snapshot: Callable[[], Any], selector):
    return use_sync_external_store(subscribe, lambda: selector(get_snapshot()))


def merge_state(d1: S, **kwargs) -> S:
    if dataclasses.is_dataclass(d1):
        return dataclasses.replace(d1, **kwargs)
    if "pydantic" in sys.modules and isinstance(d1, sys.modules["pydantic"].BaseModel):
        return type(d1)(**{**d1.dict(), **kwargs})  # type: ignore
    return cast(S, {**cast(dict, d1), **kwargs})


class ValueBase(Generic[T]):
    def __init__(self, merge: Callable = merge_state):
        self.merge = merge
        self.listeners: Set[Callable[[T], None]] = set()
        self.listeners2: Set[Callable[[T, T], None]] = set()

    @property
    def lock(self):
        raise NotImplementedError

    @property
    def value(self) -> T:
        return self.get()

    @value.setter
    def value(self, value: T):
        self.set(value)

    def set(self, value: T):
        raise NotImplementedError

    def get(self) -> T:
        raise NotImplementedError

    def subscribe(self, listener: Callable[[T], None]):
        self.listeners.add(listener)

        def cleanup():
            self.listeners.remove(listener)

        return cleanup

    def subscribe_change(self, listener: Callable[[T, T], None]):
        self.listeners2.add(listener)

        def cleanup():
            self.listeners2.remove(listener)

        return cleanup

    def update(self, _f=None, **kwargs):
        if _f is not None:
            assert not kwargs
            with self.lock:
                kwargs = _f(self.get())
        with self.lock:
            # important to have this part thread-safe
            new = self.merge(self.get(), **kwargs)
            self.set(new)

    def fire(self, new: T, old: T):
        for listener in self.listeners.copy():
            listener(new)
        for listener2 in self.listeners2.copy():
            listener2(new, old)

    def use_value(self) -> T:
        # .use with the default argument doesn't give good type inference
        return self.use()

    def use(self, selector: Callable[[T], TS] = lambda x: x) -> TS:  # type: ignore
        slice = use_sync_external_store_with_selector(
            self.subscribe,
            self.get,
            selector,
        )
        return slice

    def use_state(self) -> Tuple[T, Callable[[T], None]]:
        setter = self.set
        value = self.use()  # type: ignore
        return value, setter

    @property
    def fields(self) -> T:
        # we lie about the return type, but in combination with
        # setter we can make type safe setters (see docs/tests)
        return cast(T, Fields(self))

    def setter(self, field: TS) -> Callable[[TS], None]:
        _field = cast(FieldBase, field)

        def setter(new_value: TS):
            _field.set(new_value)

        return cast(Callable[[TS], None], setter)


# the default store for now, stores in a global dict, or when in a solara
# context, in the solara user context


class ConnectionStore(ValueBase[S]):
    _global_dict: Dict[str, S] = {}  # outside of solara context, this is used
    scope_lock = threading.Lock()

    def __init__(self, default_value: S = None, key=None):
        super().__init__()
        self.default_value = default_value
        cls = type(default_value)
        self.storage_key = key or (cls.__module__ + ":" + cls.__name__ + "-" + str(id(default_value)))
        self._global_dict = {}
        # since a set can trigger events, which can trigger new updates, we need a recursive lock
        self._lock = threading.RLock()

    @property
    def lock(self):
        return self._lock

    def _get_dict(self):
        scope_dict = self._global_dict
        if _using_solara_server():
            import solara.server.app

            try:
                context = solara.server.app.get_current_context()
            except:  # noqa
                pass  # do we need to be more strict?
            else:
                scope_dict = cast(Dict[str, S], context.user_dicts)
        return scope_dict

    def get(self):
        scope_dict = self._get_dict()
        if self.storage_key not in scope_dict:
            with self.scope_lock:
                if self.storage_key not in scope_dict:
                    # we assume immutable, so don't make a copy
                    scope_dict[self.storage_key] = self.default_value
        return scope_dict[self.storage_key]

    def set(self, value: S):
        scope_dict = self._get_dict()
        old = self.get()
        if equals(old, value):
            return
        scope_dict[self.storage_key] = value

        if _DEBUG:
            import traceback

            traceback.print_stack(limit=17, file=sys.stdout)

            print("change old", old)  # noqa
            print("change new", old)  # noqa

        self.fire(value, old)


class Reactive(ValueBase[S]):
    _storage: ValueBase[S]

    def __init__(self, default_value: Union[S, ValueBase[S]]):
        super().__init__()
        if not isinstance(default_value, ValueBase):
            self._storage = ConnectionStore(default_value, key=id(self))
        else:
            self._storage = default_value
        self.__post__init__()

    @property
    def lock(self):
        return self._storage.lock

    def __post__init__(self):
        pass

    def update(self, **kwargs):
        self._storage.update(**kwargs)

    def set(self, value: S):
        self._storage.set(value)

    def get(self) -> S:
        return self._storage.get()

    def subscribe(self, listener: Callable[[S], None]):
        return self._storage.subscribe(listener)

    def subscribe_change(self, listener: Callable[[S, S], None]):
        return self._storage.subscribe_change(listener)

    def computed(self, f: Callable[[S], T]) -> "Computed[T]":
        return Computed(f, self)


class Computed(Generic[T]):
    def __init__(self, compute: Callable[[S], T], state: Reactive[S]):
        self.compute = compute
        self.state = state

    def get(self) -> T:
        return self.compute(self.state.get())

    def subscribe(self, listener: Callable[[T], None]):
        return self.state.subscribe(lambda _: listener(self.get()))

    def use(self, selector: Callable[[T], T]) -> T:
        slice = use_sync_external_store_with_selector(
            self.subscribe,
            self.get,
            selector,
        )
        return slice


class ValueSubField(ValueBase[T]):
    def __init__(self, field: "FieldBase"):
        super().__init__()  # type: ignore
        self._field = field
        field = field
        while not isinstance(field, ValueBase):
            field = field._parent
        self._root = field
        assert isinstance(self._root, ValueBase)

    @property
    def lock(self):
        return self._root.lock

    def subscribe(self, listener: Callable[[T], None]):
        def on_change(new, old):
            new_value = self._field.get(new)
            old_value = self._field.get(old)
            if not equals(new_value, old_value):
                listener(new_value)

        return self._root.subscribe_change(on_change)

    def get(self, obj=None) -> T:
        return self._field.get(obj)

    def set(self, value: T):
        self._field.set(value)


def Ref(field: T) -> ValueSubField[T]:
    _field = cast(FieldBase, field)
    return ValueSubField[T](_field)


class FieldBase:
    _parent: Any

    def __getattr__(self, key):
        if key in ["_parent", "set", "_lock"] or key.startswith("__"):
            return self.__dict__[key]
        return FieldAttr(self, key)

    def __getitem__(self, key):
        return FieldItem(self, key)

    def get(self, obj=None):
        raise NotImplementedError

    def set(self, value):
        raise NotImplementedError


class Fields(FieldBase):
    def __init__(self, state: ValueBase):
        self._parent = state
        self._lock = state.lock

    def get(self, obj=None):
        # we are at the root, so override the object
        # so we can get the 'old' value
        if obj is not None:
            return obj
        return self._parent.get()

    def set(self, value):
        self._parent.set(value)


class FieldAttr(FieldBase):
    def __init__(self, parent, key: str):
        self._parent = parent
        self.key = key
        self._lock = parent._lock

    def get(self, obj=None):
        obj = self._parent.get(obj)
        return getattr(obj, self.key)

    def set(self, value):
        with self._lock:
            parent_value = self._parent.get()
            if isinstance(self.key, str):
                parent_value = merge_state(parent_value, **{self.key: value})
                self._parent.set(parent_value)
            else:
                raise TypeError(f"Type of key {self.key!r} is not supported")


class FieldItem(FieldBase):
    def __init__(self, parent, key: str):
        self._parent = parent
        self.key = key
        self._lock = parent._lock

    def get(self, obj=None):
        obj = self._parent.get(obj)
        return getitem(obj, self.key)

    def set(self, value):
        with self._lock:
            parent_value = self._parent.get()
            if isinstance(self.key, int) and isinstance(parent_value, (list, tuple)):
                parent_type = type(parent_value)
                parent_value = parent_value.copy()  # type: ignore
                parent_value[self.key] = value
                self._parent.set(parent_type(parent_value))
            else:
                parent_value = merge_state(parent_value, **{self.key: value})
                self._parent.set(parent_value)


# alias for compatibility
State = Reactive



import ipyvuetify as v

import solara


def test_sidebar():
    @solara.component
    def Content1():
        def on_click():
            print("click")  # noqa

        with solara.Card("Content") as content:
            with solara.Sidebar():
                solara.Button("Hi1", on_click=on_click)
            solara.Markdown("This is the content1")
        return content

    @solara.component
    def Content2():
        with solara.Card("Content") as content:
            solara.Title("Title2")
            with solara.Sidebar():
                solara.Button("Hi2")
            solara.Markdown("This is the content2")
        return content

    @solara.component
    def Sub3():
        with solara.Card("Sub3") as sub:
            solara.Markdown("Sub3 is the best")
            with solara.Sidebar():
                solara.Button("Hi3b")
        return sub

    @solara.component
    def Content3():
        with solara.Card("Content") as content:
            solara.Title("Title3")
            with solara.Sidebar():
                solara.Button("Hi3")
            solara.Markdown("This is the content3")
            Sub3()
        return content

    @solara.component
    def Content4():
        with solara.Card("Content") as content:
            solara.Title("Title4")
            solara.Markdown("This is the content4")
        return content

    set_content = None

    @solara.component
    def Layout():
        nonlocal set_content
        content, set_content = solara.use_state(1)
        with solara.AppLayout(title="Scatter plot") as main:
            if content == 1:
                Content1()
            if content == 2:
                Content2()
            if content == 3:
                Content3()
            if content == 4:
                Content4()
        return main

    box, rc = solara.render(Layout(), handle_error=False)
    assert len(rc.find(v.NavigationDrawer).find(v.Btn, children=["Hi1"])) == 1
    rc.find(v.ToolbarTitle, children=["Scatter plot"]).assert_not_empty()
    assert set_content is not None
    set_content(2)
    assert len(rc.find(v.NavigationDrawer).find(v.Btn, children=["Hi1"])) == 0
    assert len(rc.find(v.NavigationDrawer).find(v.Btn, children=["Hi2"])) == 1
    assert len(rc.find(v.ToolbarTitle, children=["Title2"])) == 1
    set_content(1)
    assert len(rc.find(v.NavigationDrawer).find(v.Btn, children=["Hi1"])) == 1
    rc.find(v.ToolbarTitle, children=["Scatter plot"]).assert_not_empty()
    set_content(3)
    assert len(rc.find(v.NavigationDrawer).find(v.Btn)) == 2
    assert len(rc.find(v.ToolbarTitle, children=["Title3"])) == 1
    set_content(4)
    rc.find(v.NavigationDrawer).assert_empty()



import os
from datetime import date, datetime, timedelta
from typing import Callable, List, Tuple, TypeVar, cast

import ipyvue
import ipyvuetify
import reacton.core
import traitlets

import solara
from solara.alias import rv

T = TypeVar("T")


@solara.value_component(int)
def SliderInt(
    label: str,
    value: int = 0,
    min: int = 0,
    max: int = 10,
    step: int = 1,
    on_value: Callable[[int], None] = None,
    thumb_label=True,
    disabled: bool = False,
):
    """Slider for controlling an integer value.

    ## Arguments

    * `label`: Label to display next to the slider.
    * `value`: The currently selected value.
    * `min`: Minimum value.
    * `max`: Maximum value.
    * `step`: Step size.
    * `on_value`: Callback to call when the value changes.
    * `thumb_label`: Show a thumb label when sliding (True), always ("always"), or never (False).
    * `disabled`: Whether the slider is disabled.
    """

    def set_value_cast(value):
        if on_value is None:
            return
        on_value(int(value))

    return rv.Slider(
        v_model=value,
        on_v_model=set_value_cast,
        label=label,
        min=min,
        max=max,
        step=step,
        thumb_label=thumb_label,
        dense=False,
        hide_details=True,
        disabled=disabled,
    )


@solara.value_component(None)
def SliderRangeInt(
    label: str,
    value: Tuple[int, int] = (1, 3),
    min: int = 0,
    max: int = 10,
    step: int = 1,
    on_value: Callable[[Tuple[int, int]], None] = None,
    thumb_label=True,
    disabled: bool = False,
) -> reacton.core.ValueElement[ipyvuetify.RangeSlider, Tuple[int, int]]:
    """Slider for controlling a range of integer values.

    ## Arguments
    * `label`: Label to display next to the slider.
    * `value`: The currently selected value.
    * `min`: Minimum value.
    * `max`: Maximum value.
    * `step`: Step size.
    * `on_value`: Callback to call when the value changes.
    * `thumb_label`: Show a thumb label when sliding (True), always ("always"), or never (False).
    * `disabled`: Whether the slider is disabled.
    """

    def set_value_cast(value):
        if on_value is None:
            return
        v1, v2 = value
        on_value((int(v1), int(v2)))

    return cast(
        reacton.core.ValueElement[ipyvuetify.RangeSlider, Tuple[int, int]],
        rv.RangeSlider(
            v_model=value,
            on_v_model=set_value_cast,
            label=label,
            min=min,
            max=max,
            step=step,
            thumb_label=thumb_label,
            dense=False,
            hide_details=True,
            disabled=disabled,
        ),
    )


@solara.value_component(float)
def SliderFloat(
    label: str,
    value: float = 0,
    min: float = 0,
    max: float = 10.0,
    step: float = 0.1,
    on_value: Callable[[float], None] = None,
    thumb_label=True,
    disabled: bool = False,
):
    """Slider for controlling a float value.

    ## Arguments
    * `label`: Label to display next to the slider.
    * `value`: The current value.
    * `min`: The minimum value.
    * `max`: The maximum value.
    * `step`: The step size.
    * `on_value`: Callback to call when the value changes.
    * `thumb_label`: Show a thumb label when sliding (True), always ("always"), or never (False).
    * `disabled`: Whether the slider is disabled.
    """

    def set_value_cast(value):
        if on_value is None:
            return
        on_value(float(value))

    return rv.Slider(
        v_model=value,
        on_v_model=set_value_cast,
        label=label,
        min=min,
        max=max,
        step=step,
        thumb_label=thumb_label,
        dense=False,
        hide_details=True,
        disabled=disabled,
    )


@solara.value_component(None)
def SliderRangeFloat(
    label: str,
    value: Tuple[float, float] = (1.0, 3.0),
    min: float = 0.0,
    max: float = 10.0,
    step: float = 0.1,
    on_value: Callable[[Tuple[float, float]], None] = None,
    thumb_label=True,
    disabled: bool = False,
) -> reacton.core.ValueElement[ipyvuetify.RangeSlider, Tuple[float, float]]:
    """Slider for controlling a range of float values.

    ## Arguments
    * `label`: Label to display next to the slider.
    * `value`: The current value.
    * `min`: The minimum value.
    * `max`: The maximum value.
    * `step`: The step size.
    * `on_value`: Callback to call when the value changes.
    * `thumb_label`: Show a thumb label when sliding (True), always ("always"), or never (False).
    * `disabled`: Whether the slider is disabled.
    """

    def set_value_cast(value):
        if on_value is None:
            return
        v1, v2 = value
        on_value((float(v1), float(v2)))

    return cast(
        reacton.core.ValueElement[ipyvuetify.RangeSlider, Tuple[float, float]],
        rv.RangeSlider(
            v_model=value,
            on_v_model=set_value_cast,
            label=label,
            min=min,
            max=max,
            step=step,
            thumb_label=thumb_label,
            dense=False,
            hide_details=True,
            disabled=disabled,
        ),
    )


@solara.value_component(None)
def SliderValue(
    label: str,
    value: T,
    values: List[T],
    on_value: Callable[[T], None] = None,
    disabled: bool = False,
) -> reacton.core.ValueElement[ipyvuetify.Slider, T]:
    """Slider for selecting a value from a list of values.

    ## Arguments
    * `label`: Label to display next to the slider.
    * `value`: The currently selected value.
    * `values`: List of values to select from.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the slider is disabled.

    """
    index, set_index = solara.use_state(values.index(value), key="index")

    def on_index(index):
        set_index(index)
        value = values[index]
        if on_value:
            on_value(value)

    return cast(
        reacton.core.ValueElement[ipyvuetify.Slider, T],
        rv.Slider(
            v_model=index,
            on_v_model=on_index,
            ticks=True,
            tick_labels=values,
            label=label,
            min=0,
            max=len(values) - 1,
            dense=False,
            hide_details=True,
            disabled=disabled,
        ),
    )


class DateSliderWidget(ipyvue.VueTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "slider_date.vue"))

    min = traitlets.CFloat(0).tag(sync=True)
    days = traitlets.CFloat(0).tag(sync=True)
    value = traitlets.Any(0).tag(sync=True)

    label = traitlets.Unicode("").tag(sync=True)
    disabled = traitlets.Bool(False).tag(sync=True)


@solara.value_component(date)
def SliderDate(
    label: str,
    value: date = date(1981, 7, 28),
    min: date = date(1981, 1, 1),
    max: date = date(3000, 12, 30),
    on_value: Callable[[date], None] = None,
    disabled: bool = False,
):
    """Slider for controlling a date value.

    ## Arguments
    * `label`: Label to display next to the slider.
    * `value`: The current value.
    * `min`: The minimum value.
    * `max`: The maximum value.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the slider is disabled.
    """

    def format(d: date):
        return float(datetime(d.year, d.month, d.day).timestamp())

    dt_min = format(min)
    delta: timedelta = max - min
    days = delta.days

    delta_value: timedelta = value - min
    days_value = delta_value.days
    if days_value < 0:
        days_value = 0

    def set_value_cast(value):
        date = min + timedelta(days=value)
        if on_value:
            on_value(date)

    return DateSliderWidget.element(label=label, min=dt_min, days=days, on_value=set_value_cast, value=days_value, disabled=disabled)


FloatSlider = SliderFloat
IntSlider = SliderInt
ValueSlider = SliderValue
DateSlider = SliderDate



import logging

# import sys
from pathlib import Path

import ipywidgets

# import pytest
import reacton.core

# import solara.server.app
from solara.server import reload
from solara.server.app import AppScript

logger = logging.getLogger("solara.server.app_test")


HERE = Path(__file__).parent
reload.reloader.start()


def test_notebook_element(app_context, no_app_context):
    name = str(HERE / "solara_test_apps" / "notebookapp_element.ipynb")
    app = AppScript(name)
    try:
        with app_context:
            el = app.run()
            assert isinstance(el, reacton.core.Element)
            el2 = app.run()
            assert el is el2
    finally:
        app.close()


def test_notebook_component(app_context, no_app_context):
    name = str(HERE / "solara_test_apps" / "notebookapp_component.ipynb")
    app = AppScript(name)
    try:
        with app_context:
            el = app.run().kwargs["children"][0].component
            assert isinstance(el, reacton.core.Component)
            el2 = app.run().kwargs["children"][0].component
            assert el is el2
    finally:
        app.close()


def test_notebook_widget(app_context, no_app_context):
    name = str(HERE / "solara_test_apps" / "notebookapp_widget.ipynb")
    app = AppScript(name)
    try:
        with app_context:
            widget = app.run()
            assert isinstance(widget, ipywidgets.Button)
            widget2 = app.run()
        assert widget is not widget2
    finally:
        app.close()


# these make other test fail on CI (vaex is used, which causes a blake3 reload, which fails)
# def test_watch_module_reload(tmpdir, app_context, extra_include_path, no_app_context):
#     import ipyvuetify as v

#     with extra_include_path(str(tmpdir)):
#         py_file = tmpdir / "test.py"
#         py_mod_file = tmpdir / "somemod.py"

#         logger.info("writing files")
#         with open(py_mod_file, "w") as f:
#             f.write("import ipyvuetify as v; App = v.Btn.element\n")
#         with open(py_file, "w") as f:
#             f.write("import somemod; app=somemod.App\n")

#         logger.info("wrote files")

#         app = AppScript(f"{py_file}")
#         try:
#             result = app.run()
#             assert "somemod" in sys.modules
#             assert "somemod" in reload.reloader.watched_modules
#             somemod1 = sys.modules["somemod"]
#             assert result().component.widget == v.Btn

#             # change depending module
#             with open(py_mod_file, "w") as f:
#                 f.write("import ipyvuetify as v; App = v.Card.element\n")
#             # wait for the event to trigger
#             reload.reloader.reload_event_next.wait()
#             # assert "somemod" not in sys.modules
#             # breakpoint()
#             result = app.run()
#             assert "somemod" in sys.modules
#             assert result().component.widget == v.Card
#             somemod2 = sys.modules["somemod"]
#             assert somemod1 is not somemod2
#         finally:
#             app.close()
#             if "somemod" in sys.modules:
#                 del sys.modules["somemod"]
#             reload.reloader.watched_modules.remove("somemod")


# def test_script_reload_component(tmpdir, app_context, extra_include_path, no_app_context):
#     import ipyvuetify as v

#     with extra_include_path(str(tmpdir)):
#         py_file = tmpdir / "test.py"

#         logger.info("writing files")
#         with open(py_file, "w") as f:
#             f.write("import reacton.ipyvuetify as v; Page = v.Btn\n")

#         app = AppScript(f"{py_file}")
#         try:
#             result = app.run()
#             assert result().component.widget == v.Btn
#             with open(py_file, "w") as f:
#                 f.write("import reacton.ipyvuetify as v; Page = v.Slider\n")
#             # wait for the event to trigger
#             reload.reloader.reload_event_next.wait()
#             # assert "somemod" not in sys.modules
#             # breakpoint()
#             result = app.run()
#             assert result().component.widget == v.Slider
#         finally:
#             app.close()


# def test_watch_module_import_error(tmpdir, app_context, extra_include_path, no_app_context):
#     import ipyvuetify as v

#     with extra_include_path(str(tmpdir)):
#         py_file = tmpdir / "test.py"
#         py_mod_file = tmpdir / "somemod2.py"

#         logger.info("writing files")
#         with open(py_mod_file, "w") as f:
#             f.write("import ipyvuetify as v; App = v.Btn.element\n")
#         with open(py_file, "w") as f:
#             f.write("import somemod2; app=somemod2.App\n")

#         logger.info("wrote files")

#         app = AppScript(f"{py_file}")
#         try:
#             result = app.run()
#             assert "somemod2" in sys.modules
#             assert "somemod2" in reload.reloader.watched_modules
#             assert result().component.widget == v.Btn

#             # syntax error
#             with open(py_mod_file, "w") as f:
#                 f.write("import ipyvuetify as v; App !%#$@= v.Card.element\n")
#             reload.reloader.reload_event_next.wait()
#             with pytest.raises(SyntaxError):
#                 result = app.run()

#             with open(py_mod_file, "w") as f:
#                 f.write("import ipyvuetify as v; App = v.Card.element\n")
#             reload.reloader.reload_event_next.wait()
#             result = app.run()
#             assert "somemod2" in sys.modules
#             assert result().component.widget == v.Card
#         finally:
#             app.close()
#             del sys.modules["somemod2"]
#             reload.reloader.watched_modules.remove("somemod2")



"""
# Overview
Click on one of the items on the left.
"""

import inspect

import solara
from solara.alias import rv

from .. import List
from .. import SimpleListItem as ListItem

_title = "API"


@solara.component
def Page():
    return solara.Markdown(__doc__)


@solara.component
def NoPage():
    raise RuntimeError("This page should not be rendered")


@solara.component
def Sidebar(children=[], level=0):
    # note that we don't use children here, but we used route.module instead to ge the module
    # this is fine because all api/*.py files use the standard Page component, and do not add
    # a new Layout component
    route_current, all_routes = solara.use_route()
    if route_current is None:
        return solara.Error("Page not found")

    # keeps track of which routes we includes
    routes = {r.path: r for r in all_routes.copy()}

    def add(path):
        route = routes[path]
        with solara.Link(route):
            ListItem(route.label, class_="active" if route_current is not None and path == route_current.path else None)
        del routes[path]

    # with solara.HBox(grow=True) as main:
    with rv.Col(tag="aside", md=4, lg=3, class_="sidebar bg-grey d-none d-md-block") as main:
        with solara.Head():
            name = route_current.label if route_current.label is not None else "No name"
            if name == "API":
                solara.Title("Solara » API overview")
            else:
                solara.Title("Solara » API » " + name)
        with List():
            add("/")
            with ListItem("Input", icon_name="mdi-chevron-left-box"):
                with List():
                    add("button")
                    add("checkbox")
                    add("input")
                    add("select")
                    add("slider")
                    add("togglebuttons")
                    add("file_browser")
                    add("file_drop")
            with ListItem("Output", icon_name="mdi-chevron-right-box"):
                with List():
                    add("markdown")
                    add("markdown_editor")
                    add("html")
                    add("image")
                    # add("code")
                    add("sql_code")
                    add("file_download")
                    add("tooltip")
            with ListItem("Status", icon_name="mdi-information"):
                with List():
                    add("success")
                    add("info")
                    add("warning")
                    add("error")
            with ListItem("Viz", icon_name="mdi-chart-histogram"):
                with List():
                    add("altair")
                    add("echarts")
                    add("matplotlib")
                    add("plotly")
                    add("plotly_express")
                #     ListItem("AltairChart")
            with ListItem("Layout", icon_name="mdi-page-layout-sidebar-left"):
                with List():
                    add("columns")
                    add("columns_responsive")
                    add("column")
                    add("row")
                    add("hbox")
                    add("vbox")
                    add("griddraggable")
                    add("gridfixed")
                    add("app_layout")
                    add("app_bar")
                    add("sidebar")
                    # add("app")
            with ListItem("Data", icon_name="mdi-database"):
                with List():
                    # ListItem("DataTable")
                    add("dataframe")
                    # add("pivot_table")
            with ListItem("Page", icon_name="mdi-file-code"):
                with List():
                    add("head")
                    add("title")
            with ListItem("Hooks", icon_name="mdi-hook"):
                with List():
                    # add("use_fetch")
                    # add("use_json")
                    add("use_cross_filter")
                    add("use_thread")
                    add("use_exception")
                    add("use_previous")
                    add("use_state")
                    add("use_state_or_update")
            with ListItem("Types", icon_name="mdi-fingerprint"):
                with List():
                    # ListItem("Action")
                    # ListItem("ColumnAction")
                    # ListItem("Route")
                    add("route")
            with ListItem("Routing", icon_name="mdi-router"):
                with List():
                    add("use_route")
                    add("resolve_path")
                    add("generate_routes")
                    add("generate_routes_directory")
                    # add("use_router")
                    # add("use_route_level")

                    # add("resolve_path")
                    # add("use_pathname")

                    # add("RenderPage")
                    # add("DefaultNavigation")
                    add("link")
            with ListItem("Utils", icon_name="mdi-hammer-wrench"):
                with List():
                    add("memoize")
                    add("display")
                    add("widget")
            with ListItem("Advanced", icon_name="mdi-head-cog-outline"):
                with List():
                    add("style")
                    add("meta")
            with ListItem("Cross filter", icon_name="mdi-filter-variant-remove"):
                with List():
                    add("cross_filter_dataframe")
                    add("cross_filter_report")
                    add("cross_filter_slider")
                    add("cross_filter_select")
        if routes:
            print(f"Routes not used: {list(routes.keys())}")  # noqa

    return main


@solara.component
def Layout(children=[]):
    route_current, all_routes = solara.use_route()
    if route_current is None:
        return solara.Error("Page not found")

    if route_current.path == "/":
        return solara.Markdown(__doc__)
    else:
        with solara.HBox(grow=True) as main:
            with solara.Padding(4):
                if route_current.module:
                    # we ignore children, and make the element again
                    WithCode(route_current.module)
        return main


@solara.component
def WithCode(module):
    # e = solara.use_exception_handler()
    # if e is not None:
    #     return solara.Error("oops")
    component = getattr(module, "Page", None)
    show_code, set_show_code = solara.use_state(False)
    with rv.Sheet() as main:
        with rv.Dialog(v_model=show_code, on_v_model=set_show_code):
            with rv.Sheet(class_="pa-4"):
                if component:
                    if hasattr(module, "sources"):
                        codes = [inspect.getsource(k) for k in module.sources]
                        code = "\n".join(codes)
                    else:
                        code = inspect.getsource(component.f)
                    code = code.replace("```", "~~~")
                    pre = ""
                    solara.MarkdownIt(
                        f"""
```python
{pre}{code}
```
"""
                    )
        # It renders code better
        solara.Markdown(
            module.__doc__ or "# no docs yet",
            unsafe_solara_execute=True,
        )
        if component and component != NoPage:
            with solara.Card("Example", margin=0, classes=["mt-8"]):
                component()
                solara.Button("Show code", icon_name="mdi-eye", on_click=lambda: set_show_code(True), text=True, class_="mt-8")

                # code = inspect.getsource(module)
                # code_quoted = urllib.parse.quote_plus(code)
                # url = f"https://test.solara.dev/try?code={code_quoted}"
                # solara.Button("Run on solara.dev", icon_name="mdi-pencil", href=url, target="_blank")
    return main



import pathlib
import shutil
import subprocess
import tempfile

from hatchling.builders.hooks.plugin.interface import BuildHookInterface

packages = [
    ["@widgetti/solara-vuetify-app", "3.0.1"],
    ["requirejs", "2.3.6"],
    ["mermaid", "8.6.4"],
    ["codemirror", "5.65.3"],
    ["vega", "5.21.0"],
    ["vega-lite", "5.2.0"],
    ["vega-embed", "6.20.2"],
    ["@widgetti/vue-grid-layout", "2.3.13-alpha.2"],
    ["@widgetti/solara-milkdown", "6.3.0"],
    ["echarts", "5.4.0"],
    ["font-awesome", "4.5.0"],
]


def npm_pack(base_cache_dir: pathlib.Path, package: str, version: str):
    with tempfile.TemporaryDirectory() as temp_dir_name:
        target_directory = base_cache_dir / f"{package}@{version}"
        if target_directory.exists():
            if (target_directory / "package.json").exists():
                print(f"package: {package} already downloaded, skipping")  # noqa
                return
            else:
                # if the directory exists and we 'move', the directory will be
                # moved to the 'package' subdirectory, so remove it first
                target_directory.unlink()
        try:
            subprocess.check_call(f"npm pack {package}@{version}", cwd=temp_dir_name, shell=True)
            package_file_name = package
            if package.startswith("@"):
                package_file_name = package[1:].replace("/", "-")
            subprocess.check_call(f"tar xzf {package_file_name}-{version}.tgz", cwd=temp_dir_name, shell=True)
            shutil.move(str(pathlib.Path(temp_dir_name) / "package"), str(target_directory))
        except Exception:
            # creating for convenience, to unpack a tarball
            # in the right directory
            target_directory.mkdir(exist_ok=True, parents=True)
            raise


def download_cdn(cache_dir):
    for package, version in packages:
        npm_pack(cache_dir, package, version)


class CustomHook(BuildHookInterface):
    def initialize(self, version, build_data):
        if self.target_name not in ["wheel", "sdist"]:
            return
        download_cdn(pathlib.Path("cdn"))



from pathlib import Path

import solara
from solara.alias import rv
from solara.components.title import Title
from solara.routing import route_level_context

from ..components import Header, Hero

directory = Path(__file__).parent

# md file
md = open(directory / "README.md").read()

title = "Home"

route_order = ["/", "docs", "api", "examples", "apps"]


@solara.component
def Page():
    with solara.GridFixed(1, justify_items="center") as main:
        solara.Markdown(md)
    return main


@solara.component
def SimpleListItem(text=None, children=[], class_: str = None, icon_name=None):
    if icon_name is not None:
        children = [rv.Icon(children=[icon_name])] + children
    if text is not None:
        children = [text] + children
    return rv.Html(tag="li", children=children, attributes={"class": class_})


@solara.component
def List(children=[], class_: str = None):
    return rv.Html(tag="ul", children=children, attributes={"class": class_})


@solara.component
def Sidebar():
    route_current, all_routes = solara.use_route()
    router = solara.use_router()
    selected = router.path
    with rv.Col(tag="aside", md=4, lg=3, class_="sidebar bg-grey d-none d-md-block") as main:
        with List():
            for route in all_routes:
                if route.children and route.data is None:
                    path = solara.resolve_path(route.children[0])
                    with solara.Link(path):
                        with SimpleListItem(route.label, class_="active" if path == selected else None):
                            with List():
                                for child in route.children[1:]:
                                    path = solara.resolve_path(child)
                                    with solara.Link(path):
                                        title = child.label or "no label"
                                        if callable(title):
                                            title = "Error: dynamic title"
                                        SimpleListItem(title, class_="active" if path == selected else None)
                else:
                    path = solara.resolve_path(route)
                    with solara.Link(path):
                        SimpleListItem(route.label, class_="active" if path == selected else None)
    return main


@solara.component
def Layout(children=[]):
    router = solara.use_router()
    route_current, all_routes = solara.use_route()
    # get child routes, and restore router level, so we can also render the route of the sidebar for the mobile view
    route_level = solara.use_context(route_level_context)
    route_sidebar_current, all_routes_sidebar = solara.use_route()
    route_level_context.provide(route_level)

    show_left_menu, set_show_left_menu = solara.use_state(False)
    show_right_menu, set_show_right_menu = solara.use_state(False)

    if route_current and route_current.path == "apps":
        return children[0]
    with solara.VBox(grow=False) as main:
        Title(title="Solara documentation")
        solara.Meta(name="twitter:card", content="summary_large_image")
        solara.Meta(name="twitter:site", content="@solara_dev")
        solara.Meta(name="twitter:image", content="https://solara.dev/static/assets/images/logo-small.png")
        solara.Meta(property="og:url", content="https://solara.dev" + router.path)
        solara.Meta(property="og:image", content="https://solara.dev/static/assets/images/logo-small.png")
        solara.Meta(property="og:type", content="website")
        Header(
            on_toggle_left_menu=lambda: set_show_left_menu(not show_left_menu),
            on_toggle_right_menu=lambda: set_show_right_menu(not show_right_menu),
        )
        if route_current is not None and route_current.path == "/":
            Hero(
                title="Build large and Scalable Web Apps for Jupyter and Production",
                sub_title="Solara helps you build powerful & scalable data apps <b>faster</b> and <b>easier</b>.",
                button_text="Quickstart",
            )

        with rv.Container(tag="section", fluid=True, ma_0=True, pa_0=True, class_="fill-height mb-8"):
            if route_current is not None and route_current.path == "/":
                description = "Use ipywidgets with Solara to build powerful and scalable web apps for Jupyter and production in Python."
                # both tags in one
                solara.Meta(name="description", property="og:description", content=description)
                solara.Meta(name="twitter:description", content=description)
                solara.Meta(property="og:title", content="Solara documentation")
                solara.Meta(name="twitter:title", content="Solara documentation")

                with rv.Row(class_="ma-2"):
                    with rv.Col(md=4, offset_md=2, sm=5, offset_sm=1):
                        solara.Markdown(
                            """
                        # What is Solara?

                        Solara lets you build web apps from pure Python using ipywidgets.

                        Grow from a one-off experiment in the Jupyter notebook to highly complex production-grade web apps with confidence.

                        Access the full power of the Python ecosystem. Use your favorite libraries.
                        """
                        )
                        with solara.HBox():
                            with solara.Link("/docs"):
                                solara.Button(label="Read more", class_="ma-1", href="/docs", color="#f19f41", dark=True)
                            with solara.Link("/docs/quickstart"):
                                solara.Button(label="Quickstart", class_="ma-1", color="#f19f41", dark=True)
                    with rv.Col(md=4, sm=5):
                        rv.Img(src="/static/public/landing/what.png", style_="width:900px")

                with rv.Row(class_="ma-8"):
                    with rv.Col(md=4, offset_md=2, sm=5, offset_sm=1):
                        rv.Img(src="/static/public/landing/complexity.png", style_="width:500px")
                    with rv.Col(md=4, sm=5):
                        solara.Markdown(
                            """
                        # Build **large** apps with **low** code complexity

                        With Solara, you can build large scale apps without hitting a complexity wall.

                        With other tools you may hit a dead end due to missing features or implementing features
                        adds too much complexity to your code base.

                        Solara offers the **flexibility** to build complex apps, but keeps the **simplicity** of a small code base.
                        """
                        )
                with rv.Row(class_="ma-8"):
                    with rv.Col(md=4, sm=5, offset_sm=1, offset_md=2):
                        solara.Markdown(
                            """
                        # The trustworthiness of React

                        Using the same API as React, but ported to Python, Solara lets you build apps with the same
                        trustworthiness as React.

                        With a decade of experience, React is battle-tested and proven to be
                        a reliable and robust framework to build large scale apps.
                        """
                        )
                    with rv.Col(md=5, sm=5):
                        rv.Img(src="/static/public/landing/python-love-react.png", style_="width:300px")
                with rv.Row(class_="ma-8"):
                    with rv.Col(md=5, offset_md=2, sm=5, offset_sm=1):
                        solara.Markdown(
                            """
                        ## For any consulting, training or support needs
                        [contact@solara.dev](mailto:contact@solara.dev)
                        """
                        )
            else:
                with rv.Row(style_="gap:6rem; flex-wrap: nowrap;"):
                    if route_current is not None and hasattr(route_current.module, "Sidebar"):
                        route_current.module.Sidebar()  # type: ignore
                    else:
                        Sidebar()
                    with rv.Col(tag="main", md=True, class_="pt-12 pl-12 pr-10", style_="max-width: 1024px; overflow: auto;"):
                        if route_current is not None and route_current.path == "/":
                            with rv.Row(align="center"):
                                pass
                                # with rv.Col(md=6, class_="pa-0"):
                                #     rv.Html(tag="h1", children=["Live Demo"])
                                # with rv.Col(md=6, class_="d-flex", style_="justify-content: end"):
                                #     rv.Btn(elevation=0, large=True, children=["Running App"], color="primary", class_="btn-size--xlarge")
                            # solara.Padding(6)
                        with rv.Row(children=children, class_="solara-page-content-search"):
                            pass

            # Drawer navigation for top menu
            with rv.NavigationDrawer(
                v_model=show_right_menu,
                on_v_model=set_show_right_menu,
                fixed=True,
                right=True,
                hide_overlay=False,
                overlay_color="#000000",
                overlay_opacity=0.5,
                style_="height: 100vh",
            ):
                with rv.List(nav=True):
                    with rv.ListItemGroup(active_class="text--primary"):
                        for route in all_routes:
                            if route.path == "apps":
                                continue
                            with solara.Link(route):
                                solara.ListItem(route.label)

            # Drawer navigation for sidebar
            with rv.NavigationDrawer(
                v_model=show_left_menu,
                on_v_model=set_show_left_menu,
                fixed=True,
                hide_overlay=False,
                overlay_color="#000000",
                overlay_opacity=0.5,
                style_="height: 100vh",
            ):
                with rv.List(nav=True):
                    current_path = router.path
                    with rv.ListItemGroup(active_class="text--primary", v_model=current_path):
                        for route in all_routes_sidebar:
                            # this gets rid of the api/link child routes
                            if len([k for k in route.children if k.label]) == 0:
                                with solara.Link(route):
                                    solara.ListItem(route.label, value=solara.resolve_path(route))
                            else:
                                with solara.ListItem(route.label or "no-title", value=solara.resolve_path(route) + "_no_select"):
                                    for subroute in route.children:
                                        if subroute.label:
                                            with solara.Link(subroute):
                                                with solara.ListItem(subroute.label, value=solara.resolve_path(subroute)):
                                                    pass

    return main



import solara


@solara.component
def Page():
    with solara.Card("Fullscreen apps"):
        solara.Markdown(
            """
        These examples need to be run in the full browser window

         * [Streamlit users tutorial](/apps/tutorial-streamlit)
         * [Scatter plot app](/apps/scatter)
         * [Multipage](/apps/multipage)
         * [Layout demo](/apps/layout-demo)

        """
        )



import logging
import pathlib

import requests

from . import settings

logger = logging.getLogger("Solara.cdn")

cdn_url_path = "_solara/cdn"


def put_in_cache(base_cache_dir: pathlib.Path, path, data: bytes):
    cache_path = base_cache_dir / path
    pathlib.Path(cache_path.parent).mkdir(parents=True, exist_ok=True)
    try:
        logger.info("Writing cache file: %s", cache_path)
        cache_path.write_bytes(data)
    except FileNotFoundError:
        logger.info("Failed writing cache file: %s", cache_path)


def get_from_cache(base_cache_dir: pathlib.Path, path):
    cache_path = base_cache_dir / path
    try:
        logger.info("Opening cache file: %s", cache_path)
        return cache_path.read_bytes()
    except FileNotFoundError:
        pass


def get_cdn_url(path):
    path = str(path)  # on windows, the path can contain a \
    return str(settings.assets.cdn) + str(path).replace("\\", "/")


def get_data(base_cache_dir: pathlib.Path, path):
    parts = path.replace("\\", "/").split("/")
    store_path = path if len(parts) != 1 else pathlib.Path(path) / "__main.js"

    content = get_from_cache(base_cache_dir, store_path)
    if content:
        return content

    url = get_cdn_url(path)
    response = requests.get(url)
    if response.ok:
        put_in_cache(base_cache_dir, store_path, response.content)
        return response.content
    else:
        logger.warning("Could not load URL: %r", url)
        raise Exception(f"Could not load URL: {url}")


def get_path(base_cache_dir: pathlib.Path, path) -> pathlib.Path:

    parts = path.replace("\\", "/").split("/")
    store_path = path if len(parts) != 1 else pathlib.Path(path) / "__main.js"
    cache_path = base_cache_dir / store_path

    if cache_path.exists():
        return cache_path
    url = get_cdn_url(path)
    response = requests.get(url)
    if response.ok:
        put_in_cache(base_cache_dir, store_path, response.content)
        assert cache_path.exists(), f"Could not write to {cache_path}"
        return cache_path
    else:
        logger.warning("Could not load URL: %r", url)
        raise Exception(f"Could not load URL: {url}")



"""
# ToggleButtons

ToggleButtons are in two flavours, for single, and for multiple selections.


"""
import solara
from solara.website.utils import apidoc

title = "ToggleButtons"


@solara.component
def Page():
    with solara.VBox() as main:
        with solara.Card("Single selection"):
            food, set_food = solara.use_state("Banana")
            solara.Markdown(f"**Selected**: {food}")
            solara.ToggleButtonsSingle(food, values=["Kiwi", "Banana", "Apple"], on_value=set_food)

        with solara.Card("Multiple selections"):
            all_languages = "Python C++ Java JavaScript TypeScript BASIC".split()
            languages, set_languages = solara.use_state([all_languages[0]])
            solara.Markdown(f"**Selected**: {languages}")
            solara.ToggleButtonsMultiple(languages, values=all_languages, on_value=set_languages)

        with solara.Card("Custom buttons"):
            direction, set_direction = solara.use_state("left")
            solara.Markdown(f"**Selected**: {direction}")
            # instead of using the values argument, we can use the buttons as children
            # the label of the button will be used as value, if no value is given.
            with solara.ToggleButtonsSingle(direction, on_value=set_direction):
                # note that the label and the value are different
                solara.Button("Up", icon_name="mdi-arrow-up-bold", value="up", text=True)
                solara.Button("Down", icon_name="mdi-arrow-down-bold", value="down", text=True)
                solara.Button("Left", icon_name="mdi-arrow-left-bold", value="left", text=True)
                solara.Button("Right", icon_name="mdi-arrow-right-bold", value="right", text=True)

    return main


__doc__ += "# ToggleButtonsSingle"
__doc__ += apidoc(solara.ToggleButtonsSingle.f)  # type: ignore
__doc__ += "# ToggleButtonsMultiple"
__doc__ += apidoc(solara.ToggleButtonsMultiple.f)  # type: ignore



import hashlib

import ipyvuetify as v

import solara


@solara.component
def Style(value: str = ""):
    """Add a custom piece of CSS.

    Note that this is considered an advanced feature, and should be used with caution.

    ## Arguments

    - `value`: The CSS string to insert into the page.
    """
    hash = hashlib.sha256(value.encode("utf-8")).hexdigest()
    id = hash
    # ipyvue does not remove the css itself
    script = (
        """
<script>
module.exports = {
  destroyed() {
    document.getElementById("ipyvue-%s").remove();
  },
};
</script>
    """
        % id
    )

    template = f"""
<template>
    <span style="display: none">
    </span>
</template>
{script}
<style id="{id}">
{value}
</style>
    """
    # using .key avoids re-using the template, which causes a flicker (due to ipyvue)
    return v.VuetifyTemplate.element(template=template).key(id)



from typing import Any, Callable, cast

import ipyvue
import reacton

import solara
from solara.alias import rv as v


def use_change(el: reacton.core.Element, on_value: Callable[[Any], Any], enabled=True):
    """Trigger a callback when a blur events occurs or the enter key is pressed."""
    on_value_ref = solara.use_ref(on_value)
    on_value_ref.current = on_value

    def add_events():
        def on_change(widget, event, data):
            if enabled:
                on_value_ref.current(widget.v_model)

        widget = cast(ipyvue.VueWidget, solara.get_widget(el))
        if enabled:
            widget.on_event("blur", on_change)
            widget.on_event("keyup.enter", on_change)

        def cleanup():
            if enabled:
                widget.on_event("blur", on_change, remove=True)
                widget.on_event("keyup.enter", on_change, remove=True)

        return cleanup

    solara.use_effect(add_events, [enabled])


@solara.component
def InputText(
    label: str,
    value: str = "",
    on_value: Callable[[str], None] = None,
    disabled: bool = False,
    password: bool = False,
    continuous_update: bool = False,
):
    """Free form text input.

    ## Arguments

    * `label`: Label to display next to the slider.
    * `value`: The currently entered value.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the input is disabled.
    * `password`: Whether the input is a password input (typically shows input text obscured with an asterisk).
    * `continuous_update`: Whether to call the `on_value` callback on every change or only when the input loses focus or the enter key is pressed.
    """

    def set_value_cast(value):
        if on_value is None:
            return
        on_value(str(value))

    def on_v_model(value):
        if continuous_update:
            set_value_cast(value)

    text_field = v.TextField(v_model=value, on_v_model=on_v_model, label=label, disabled=disabled, type="password" if password else None)
    use_change(text_field, set_value_cast, enabled=not continuous_update)
    return text_field


@solara.component
def InputFloat(
    label: str,
    value: float = 0.0,
    on_value: Callable[[float], None] = None,
    disabled: bool = False,
    continuous_update: bool = False,
):
    """Numeric input.

    ## Arguments

    * `label`: Label to display next to the slider.
    * `value`: The currently entered value.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the input is disabled.
    * `continuous_update`: Whether to call the `on_value` callback on every change or only when the input loses focus or the enter key is pressed.
    """

    def set_value_cast(value):
        if on_value is None:
            return
        try:
            float_value = float(value)
        except Exception:
            # TODO: maybe we should show an error message here?
            return
        on_value(float_value)

    def on_v_model(value):
        if continuous_update:
            set_value_cast(value)

    text_field = v.TextField(v_model=value, on_v_model=on_v_model, label=label, disabled=disabled, type="number")
    use_change(text_field, set_value_cast, enabled=not continuous_update)
    return text_field



# isort: skip_file

from .components import *  # noqa: F401, F403
from .toestand import Reactive, Ref, State  # noqa: F401


def __getattr__(name):
    # for backwards compatibility
    from solara.components.cross_filter import (  # noqa: F401
        CrossFilterDataFrame,
        CrossFilterReport,
        CrossFilterSelect,
        CrossFilterSlider,
    )

    if name == "CrossFilterDataFrame":
        return CrossFilterDataFrame
    elif name == "CrossFilterReport":
        return CrossFilterReport
    elif name == "CrossFilterSelect":
        return CrossFilterSelect
    elif name == "CrossFilterSlider":
        return CrossFilterSlider
    else:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



import sys
import threading
from typing import MutableMapping

from .types import ObservableMutableMapping


def _in_solara_server():
    return "solara.server" in sys.modules


class ConnectionScope(ObservableMutableMapping):
    def __init__(self, name="connection"):
        super().__init__()
        self._global_dict = {}
        self.name = name
        self.lock = threading.Lock()

    def _get_dict(self) -> MutableMapping:
        if _in_solara_server():
            import solara.server.app

            context = solara.server.app.get_current_context()
            if self.name not in context.user_dicts:
                with self.lock:
                    if self.name not in context.user_dicts:
                        context.user_dicts[self.name] = {}
            return context.user_dicts[self.name]
        else:
            return self._global_dict


class ObservableDict(ObservableMutableMapping):
    def __init__(self):
        super().__init__()
        self._global_dict = {}

    def _get_dict(self) -> MutableMapping:
        return self._global_dict


worker = ObservableDict()
connection = ConnectionScope()



import ipywidgets as widgets

from solara.server import app, kernel


def test_widget_dict():
    kernel_shared = kernel.Kernel()
    context1 = app.AppContext(id="1", kernel=kernel_shared, control_sockets=[], widgets={}, templates={})
    context2 = app.AppContext(id="2", kernel=kernel_shared, control_sockets=[], widgets={}, templates={})

    with context1:
        btn1 = widgets.Button(description="context1")
    with context2:
        btn2 = widgets.Button(description="context2")

    assert btn1.model_id in context1.widgets
    assert btn1.model_id not in context2.widgets
    assert btn2.model_id in context2.widgets
    assert btn2.model_id not in context1.widgets

    del btn1
    del btn2
    context1.close()
    context2.close()
